////////////////////////////////////////////////////////////////////////////////////////////////////
//
//  Project:  Embedded Learning Library (ELL)
//  File:     MapLoadArguments.cpp (common)
//  Authors:  Ofer Dekel, Chuck Jacobs
//
////////////////////////////////////////////////////////////////////////////////////////////////////

#include "MapLoadArguments.h"

#include <model/include/Model.h>
#include <model/include/Node.h>
#include <model/include/PortElements.h>

#include <utilities/include/Files.h>

#include <sstream>

namespace ell
{
namespace common
{
    //
    // MapLoadArguments
    //
    void ParsedMapLoadArguments::AddArgs(utilities::CommandLineParser& parser)
    {
        parser.AddOption(
            inputMapFilename,
            "inputMapFilename",
            "imap",
            "Path to the input map file",
            "");

        parser.AddOption(
            inputModelFilename,
            "inputModelFilename",
            "imf",
            "Path to the input model file",
            "");

        parser.AddOption(
            modelInputsString,
            "modelInputs",
            "in",
            "Model inputs to use",
            "");

        parser.AddOption(
            modelOutputsString,
            "modelOutputs",
            "out",
            "Model outputs to use",
            "");

        parser.AddOption(
            defaultInputSize,
            "defaultInputSize",
            "d",
            "Default size of input node",
            1);
    }

    std::string MapLoadArguments::GetInputFilename() const
    {
        if (HasMapFilename())
        {
            return inputMapFilename;
        }
        else if (HasModelFilename())
        {
            return inputModelFilename;
        }
        else
        {
            return "";
        }
    }

    model::InputNodeBase* MapLoadArguments::GetInput(model::Model& model) const
    {
        // name of input node
        auto inputNode = model.GetNode(utilities::UniqueId(modelInputsString));
        return dynamic_cast<model::InputNodeBase*>(inputNode);
    }

    const model::OutputPortBase* MapLoadArguments::GetOutput(model::Model& model) const
    {
        auto elementsProxy = model::ParsePortElementsProxy(modelOutputsString);
        auto portElements = model::ProxyToPortElements(model, elementsProxy);
        return &model.SimplifyOutputs(portElements);
    }

    //
    // ParsedMapLoadArguments
    //
    utilities::CommandLineParseResult ParsedMapLoadArguments::PostProcess(const utilities::CommandLineParser& parser)
    {
        std::vector<std::string> parseErrorMessages;

        if (inputModelFilename != "" && inputMapFilename != "")
        {
            parseErrorMessages.push_back("Can only specify an input map or model file, not both.");
        }
        else if (inputMapFilename != "")
        {
            if (!utilities::IsFileReadable(inputMapFilename))
            {
                parseErrorMessages.push_back("Cannot read from specified input map file: " + inputMapFilename);
            }
        }
        else if (inputModelFilename != "")
        {
            if (!utilities::IsFileReadable(inputModelFilename))
            {
                parseErrorMessages.push_back("Cannot read from specified input model file: " + inputModelFilename);
            }
        }

        return parseErrorMessages;
    }
} // namespace common
} // namespace ell
