////////////////////////////////////////////////////////////////////////////////////////////////////
//
//  Project:  Embedded Learning Library (ELL)
//  File:     IncrementalEvaluator.h (evaluators)
//  Authors:  Ofer Dekel
//
////////////////////////////////////////////////////////////////////////////////////////////////////

#pragma once

#include "Evaluator.h"

#include <memory>
#include <vector>

namespace ell
{
namespace evaluators
{
    /// <summary> Interface to an incremental evaluator (used to evaluate ensembles). </summary>
    ///
    /// <typeparam name="BasePredictorType"> The predictor type. </typeparam>
    template <typename BasePredictorType>
    class IIncrementalEvaluator
    {
    public:
        virtual ~IIncrementalEvaluator() = default;

        /// <summary>
        /// Runs the given predictor on the evaluation set, invokes each of the aggregators on the output,
        /// and logs the result.
        /// </summary>
        ///
        /// <param name="basePredictor"> The base predictor. </param>
        /// <param name="basePredictorWeight"> The weight of the base predictor in the ensemble. </param>
        /// <param name="evaluationRescale"> A rescaling coefficient applied to the current predictions of the entire ensemble, but not recorded in the evaluator. </param>
        virtual void IncrementalEvaluate(const BasePredictorType& basePredictor, double basePredictorWeight = 1.0, double evaluationRescale = 1.0) = 0;

        /// <summary> Gets the goodness of the most recent evaluation, according to the first aggregator. </summary>
        ///
        /// <returns> The goodness of the most recent evaluation. </returns>
        virtual double GetGoodness() const = 0;

        /// <summary> Prints the logged evaluations to an output stream. </summary>
        ///
        /// <param name="os"> [in,out] The output stream. </param>
        virtual void Print(std::ostream& os) const = 0;
    };

    template <typename BasePredictorType, typename... AggregatorTypes>
    class IncrementalEvaluator : public Evaluator<BasePredictorType, AggregatorTypes...>
        , public IIncrementalEvaluator<BasePredictorType>
    {
    public:
        using BaseClassType = Evaluator<BasePredictorType, AggregatorTypes...>;

        /// <summary> Constructs an instance of IncrementalEvaluator with a given data set and given aggregators. </summary>
        ///
        /// <param name="anyDataset"> An dataset. </param>
        /// <param name="evaluatorParameters"> The evaluation parameters. </param>
        /// <param name="aggregators"> The aggregators. </param>
        IncrementalEvaluator(const data::AnyDataset& anyDataset, const EvaluatorParameters& evaluatorParameters, AggregatorTypes... aggregators);

        /// <summary>
        /// Runs the given predictor on the evaluation set, increments cached outputs, invokes each of
        /// the aggregators, and logs the result.
        /// </summary>
        ///
        /// <param name="basePredictor"> The base predictor. </param>
        /// <param name="basePredictorWeight"> The weight of the base predictor in the ensemble. </param>
        /// <param name="evaluationRescale"> A rescaling coefficient applied to the current predictions of the entire ensemble, but not recorded in the evaluator. </param>
        void IncrementalEvaluate(const BasePredictorType& basePredictor, double basePredictorWeight = 1.0, double evaluationRescale = 1.0) override;

        /// <summary> Gets the goodness of the most recent evaluation, according to the first aggregator. </summary>
        ///
        /// <returns> The goodness of the most recent evaluation. </returns>
        double GetGoodness() const override;

        /// <summary> Prints the logged evaluations to an output stream. </summary>
        ///
        /// <param name="os"> [in,out] The output stream. </param>
        void Print(std::ostream& os) const override;

    private:
        std::vector<double> _predictions;
    };

    /// <summary> Makes an incremental evaluator (used to evaluate ensembles). </summary>
    ///
    /// <typeparam name="BasePredictorType"> The predictor type. </typeparam>
    /// <typeparam name="AggregatorTypes"> The Aggregator types. </typeparam>
    /// <param name="exampleIterator"> An example iterator that represents the evaluation data. </param>
    /// <param name="aggregators"> The aggregators. </param>
    ///
    /// <returns> A unique_ptr to an IEvaluator. </returns>
    template <typename BasePredictorType, typename... AggregatorTypes>
    std::shared_ptr<IIncrementalEvaluator<BasePredictorType>> MakeIncrementalEvaluator(data::ExampleIterator<data::AutoSupervisedExample> exampleIterator, const EvaluatorParameters& evaluatorParameters, AggregatorTypes... aggregators);
} // namespace evaluators
} // namespace ell

#pragma region implementation

namespace ell
{
namespace evaluators
{
    template <typename BasePredictorType, typename... AggregatorTypes>
    IncrementalEvaluator<BasePredictorType, AggregatorTypes...>::IncrementalEvaluator(const data::AnyDataset& anyDataset, const EvaluatorParameters& evaluatorParameters, AggregatorTypes... aggregators) :
        Evaluator<BasePredictorType, AggregatorTypes...>(anyDataset, evaluatorParameters, aggregators...)
    {
        _predictions.resize(BaseClassType::_dataset.NumExamples());
    }

    template <typename BasePredictorType, typename... AggregatorTypes>
    void IncrementalEvaluator<BasePredictorType, AggregatorTypes...>::IncrementalEvaluate(const BasePredictorType& basePredictor, double basePredictorWeight, double evaluationRescale)
    {
        ++BaseClassType::_evaluateCounter;
        bool evaluate = BaseClassType::_evaluateCounter % BaseClassType::_evaluatorParameters.evaluationFrequency == 0 ? true : false;

        auto iterator = BaseClassType::_dataset.GetExampleIterator();
        size_t index = 0;

        while (iterator.IsValid())
        {
            const auto& example = iterator.Get();

            double exampleWeight = example.GetMetadata().weight;
            double label = example.GetMetadata().label;
            _predictions[index] += basePredictorWeight * basePredictor.Predict(example.GetDataVector());

            if (evaluate)
            {
                BaseClassType::DispatchUpdate(_predictions[index] * evaluationRescale, label, exampleWeight, std::make_index_sequence<sizeof...(AggregatorTypes)>());
            }

            iterator.Next();
            ++index;
        }
        if (evaluate)
        {
            BaseClassType::Aggregate(std::make_index_sequence<sizeof...(AggregatorTypes)>());
        }
    }

    template <typename BasePredictorType, typename... AggregatorTypes>
    double IncrementalEvaluator<BasePredictorType, AggregatorTypes...>::GetGoodness() const
    {
        return BaseClassType::GetGoodness();
    }

    template <typename BasePredictorType, typename... AggregatorTypes>
    void IncrementalEvaluator<BasePredictorType, AggregatorTypes...>::Print(std::ostream& os) const
    {
        BaseClassType::Print(os);
    }

    template <typename BasePredictorType, typename... AggregatorTypes>
    std::shared_ptr<IIncrementalEvaluator<BasePredictorType>> MakeIncrementalEvaluator(data::ExampleIterator<data::AutoSupervisedExample> exampleIterator, const EvaluatorParameters& evaluatorParameters, AggregatorTypes... aggregators)
    {
        return std::make_unique<IncrementalEvaluator<BasePredictorType, AggregatorTypes...>>(exampleIterator, evaluatorParameters, aggregators...);
    }
} // namespace evaluators
} // namespace ell

#pragma endregion implementation
